//Geometric types for Postgis
//
// point3d, line3d, and polygon3d are the base types.
// 
// Everything is stored in a geometry3d, which is just a conglomeration
// of the base types (and a little bit of other info).



#include "utils/geo_decls.h"



#define	POINTTYPE	1
#define	LINETYPE	2
#define	POLYGONTYPE	3
#define	MULTIPOINTTYPE	4
#define	MULTILINETYPE	5
#define	MULTIPOLYGONTYPE	6
#define	COLLECTIONTYPE	7
#define	BBOXONLYTYPE	99

/*---------------------------------------------------------------------
 * POINT3D - (x,y,z)
 *            Base type for all geometries
 *		  Also used for POINT type and MULTIPOINT type
 *-------------------------------------------------------------------*/
typedef struct
{
	double		x,y,z;
} POINT3D;

/*---------------------------------------------------------------------
 * BOX3D	- Specified by two corner points, which are
 *		 sorted to save calculation time later.
 *
 *           LLB -- lower right bottom point (ie. South West, Z low)
 *           URT -- upper right top point    (ie. North East, Z high)
 *
 *           this is the long diagonal
 *
 * example:
 *   'BOX([0.0,0.0,0.0],[10.0,10.0,10.0])'
 *
 *	NOTE:  You CAN make columns in your database of this type
 *		 IT IS NOT A TRUE GEOMETRY!
 *-------------------------------------------------------------------*/
typedef struct
{
	POINT3D		LLB,URT; /* corner POINT3Ds on long diagonal */
} BOX3D;




/*---------------------------------------------------------------------
 * LINE  - set of points (directed arcs)
 *		P1->P2->P3->P4...->Pn
 *-------------------------------------------------------------------*/
typedef struct
{
	int32 	npoints; // how many points in the line
	int32 	junk;	   // double-word alignment	
	POINT3D  	points[1]; // array of actual points
} LINE3D;

/*---------------------------------------------------------------------
 * POLYGON  - set of closed rings.  First ring in outer boundary
 *						other rings are "holes"
 *
 * NOTE: this is acually stored a bit differently.  It is more like:
 *		int32 nrings
 *		char	filler[]
 *
 *  where the 1st 4 byes of filler[] is npoints[0] the next 4 bytes
 *  are npoints[1], etc...
 *
 *  points[0] is either at filler[ 4 * nrings] 
 *            or at filler [ 4* nrings + 4]
 *		  Which ever one is double-word-aligned
 *-------------------------------------------------------------------*/
typedef struct
{
	int32 	nrings;	 // how many rings in this polygon
	int32		npoints[1]; //how many points in each ring
	/* could be 4 byes of filler here to make sure points[] is 
         double-word aligned*/
	POINT3D  	points[1]; // array of actual points
} POLYGON3D;


/*---------------------------------------------------------------------
 * Geometry - all columns in postgres are of this type
 *
 *		Geometries are collections of simple geometry types
 *		(point, line, polygon).
 *	
 *		A Point is a geometry with a single point in it.
 *		A MultiPoint is a geometry with a list of 'point' in it.
 *		A Line is a geometry with a single line in it.
 *		A MultiLine is a geometry with a list of 'line' in it.
 *		A Polygon is a geometry with a single polygon in it.
 *		A MultiPolygon is a geometry with a list of 'polygon' in it.
 *		A Collection is a geometry with a (mixed) list of 
 *				point, line, and polygon.
 *
 * 		The bvol is the bounding volume of all the subobjects.
 *		
 *		is3d is true if the original data was sent in a 3d data.
 *		2d data is 3d data with z=0.0.
 *
 *	'type' has values:
 *		Point			-> POINTTYPE
 *		MultiPoint		-> MULTIPOINTTYPE
 * 		Line			-> LINETYPE
 * 		MultiLine		-> MULTILINETYPE
 *		Polygon		-> POLYGONTYPE
 *		MultiPolygon	-> MULTIPOLYGONTYPE
 *		Collection		-> COLLECTIOMTYPE
 *
 *	'objType' has values:
 *		Point			-> POINTTYPE
 * 		Line			-> LINETYPE
 *		Polygon		-> POLYGONTYPE
 *
 *	'objOffset' is an offset (in bytes) into this structure of where
 *	the subobject is defined.  THESE ARE ALWAYS DOUBLE-WORD ALIGNED.
 *
 *
 *	In reality the structure looks like:
 *		int32 	size;
 *		int32		type;	
 * 		bool		is3d;
 *		<there is almost certainly some type of padding here>
 *		BOX3D		bvol;	
 *		int32		nobjs;
 *		char		data[...];
 *
 * 	AND:
 *		&objType[0] = &data[0]
 *		&objType[1] = &data[4] 
 *			...
 *		&obgOffset[0] = &data[ 4* nobjs]
 *		&obgOffset[1] = &data[ 4* nobjs + 4]
 *			...
 *		&objData[0] = &GEOMETRY + objOffset[0]  //always double-word aligned
 *		&objData[1] = &GEOMETRY + objOffset[1]  //always double-word aligned
 *			...	
 *
 * 	ALL GEOMETRY COLUMNS IN YOUR DATABASE SHOULD BE OF THIS TYPE
 *-------------------------------------------------------------------*/
typedef struct
{
	int32		size;		// postgres variable-length type requirement
	int32		type;		// this type of geometry
	bool		is3d;		// true if the points are 3d (only for output)
	BOX3D		bvol;		// bounding volume of all the geo objects
	int32		nobjs;	// how many sub-objects in this object
	int32		objType[1];	// type of object
	int32		objOffset[1];// offset (in bytes) into this structure where
					 // the object is located
	char		objData[1];  // store for actual objects

} GEOMETRY;


//for GiST indexing

//This is the BOX type from geo_decls.h
// Its included here for the GiST index. 
//  Originally, we used BOXONLYTYPE geometries as our keys, but after
//   Oleg and teodor (http://www.sai.msu.su/~megera/postgres/gist/)
//   have released a more generic rtree/gist index for geo_decls.h polygon
//   type.  I'm using a slightly modified version of this, so 
//   it will be easier to maintain.
//
//   Their indexing is based on the BOX object, so we include it here.


// ONLY FOR INDEXING
typedef struct geomkey {
	int32 size; /* size in varlena terms */ 
	BOX	key;
} GEOMETRYKEY; 




